! Fortran 90 and PVM
! measure of communication performance
! Ping sends a message to Pong, which sends it back to Ping.
! The measured overall time is then halved.
!
! Laurent Gasser (lga@sma.ch)
! Swiss Meteorological Institute
! November 1, 1996

program Ping

use PVM_Constants
use PVM_Interfaces
use politeness
use parameters

implicit none

  integer, parameter :: nb_echos = 1
  integer       :: Pong_Task_ID
  integer       :: my_Task_ID
  integer       :: chrono_latency      = -666
  integer       :: ticks_per_second    = -666
  integer       :: largest_packet_size = -1

!body Ping
  print *, " Ping-Pong"
  print *, " A simple measure of message passing performance"
  print *, " Laurent Gasser (lga@sma.ch)"
  print *, " Swiss Meteorological Institute, Zuerich"
  print *, " November 1, 1996"
  print *

  call Specify_Experiment()
  call Configure_Experiment()
  call Collect_Measures()
  call Report_Performances()

contains

  subroutine Configure_Experiment()
    use metrics
    use PVM_Constants
    use protocol
    implicit none
  ! local declarations
    integer              :: i          = -1
    integer              :: ignore     = -666
    integer              :: info       = PvmOK
    integer, allocatable :: items(:)
    type(durations)      :: measures   = durations(-666, -666)
    type(pvmhostinfo)    :: ping_host
    character(41)        :: slave_name = "Pong"
    integer              :: started    = -1

  ! body Configure_Experiment
    print *, " Configuring the experiment."
    print *
    ! start the correspondent Pong
    call pvm_mytid(my_Task_ID)
    call pvm_spawn(slave_name, PvmTaskHost, host_name, Pong_Task_ID, info)
    if (info /= PvmOK) then
       print *, " Pong's task failed to be spawn. "
       call pvm_perror("Ping, Configure_Experiment", info)
    end if

    ! measure the time delay of the chronograph itself
    ! in-between early and late is one complete call
    call system_clock(count = measures%early)
    call system_clock(count = measures%late)
    chrono_latency = measures%late - measures%early
    call system_clock(count_rate = ticks_per_second)
    print *, " The chronograph has a latency of ", chrono_latency, " ticks."
    print *, " They are ", ticks_per_second, " ticks per seconds."
    print *

    ! find the largest package 
    largest_packet_size = PROLOGUE_PACKET_SIZE
    do i = 1, nb_measures
       if (largest_packet_size < measure(i)%items) then
          largest_packet_size = measure(i)%items
       end if
    end do
    
    ! wait for completion of Pong's start-up
    print *, " Waiting for Pong's greeting to come..."
    call pvm_config(ignore, ignore, ping_host, info)
    if (info < PvmOK) then 
       print *, " Unable to identify configuration: error = ", info
       print ("('  Ping''s Task_ID = ', Z6)"), my_Task_ID
    else
       print ("('  Ping''s Task_ID = ', Z6, ' hosted by ', A)"), &
              my_Task_ID, trim(ping_host%hi_name)
    end if
    print ("('  Pong''s Task_ID = ', Z6, ' hosted by ', A, /)"), &
           Pong_Task_ID, trim(host_name)
    call hear_hello(from = Pong_Task_ID)

    ! transmit the number of measures to Pong
    print *, " Transfer the measures specifications to Pong."
    print *
    allocate(items(nb_measures), STAT = info)
    if (info /= 0) then
       print *, " Could not allocate an array for items."
       call pvm_perror("Ping, Configure_Experiment", info)
    end if
    do i = 1, nb_measures
       items(i) = measure(i)%items
    end do
    call specify(Pong_Task_ID, items)
  end subroutine Configure_Experiment

  subroutine Collect_Measures()
    use protocol    

  ! body Collect_Measures
    print *, " Collecting the measures."
    print *

    print *, " Packages of integers..."
    print *
    call Measure_Integers(largest_packet_size)

    print *, " Packages of reals..."
    print *
    call Measure_Reals(largest_packet_size)

    print *, " Packages of complexes..."
    print *
    call Measure_Complexes(largest_packet_size)

    print *, " Packages of logicals..."
    print *
    call Measure_Logicals(largest_packet_size)

    print *, " Packages of characters..."
    print *
    call Measure_Characters(largest_packet_size)

    print *, " Wait goodbye from Pong."
    print ("('  Pong''s Task_ID = ', Z6)"), Pong_Task_ID
    print *
    call hear_goodbye(from = Pong_Task_ID)
  end subroutine Collect_Measures

  subroutine  Report_Performances()
    use metrics
    use parameters
    implicit none
    ! local declarations
    character(25), parameter :: data_name(F90_basic_types) = (/ &
                                "integers (default kind). ", &
                                "reals (default kind).    ", &
                                "complexes (default kind).", &
                                "logicals.                ", &
                                "characters.              " /)
    integer :: data_type    = -666
    integer :: error_status = -666
    real    :: estimators(2, 2)
    integer :: experiment   = -666
    integer :: x(nb_measures)
    integer :: y(nb_measures, 2)

  ! body Report_Performances

    by_type: do data_type = 1, F90_basic_types
       print *, " Trip Time in ticks for Packets of ", data_name(data_type)
       print *
       print *, " Items    Transfer   w packing"
       raw_data: do experiment = 1, nb_measures
          print ("(I6,I12,I12)"), &
               measure(experiment)%items, &
               measure(experiment)%duration(data_type)%early, &
               measure(experiment)%duration(data_type)%late
       end do raw_data
       print *
       ! estimate the linear rule by least square
       do experiment = 1, nb_measures
          x(experiment)    = measure(experiment)%items
          ! transfer only
          y(experiment, 1) = measure(experiment)%duration(data_type)%early
          ! with packing
          y(experiment, 2) = measure(experiment)%duration(data_type)%late
       end do
       estimators(1,:) = Linear_Regression(x, y(:,1))
       estimators(2,:) = Linear_Regression(x, y(:,2))
       print ('(A F12.1 F12.1 A)'), "Latency ", &
            estimators(1,1), estimators(2,1), " [ticks]"
       print ('(A F12.6 F12.6 A)'), "Rate    ", &
            estimators(1,2), estimators(2,2), " [ticks/item]"
       print *
       print *
    end do by_type
    call pvm_exit(error_status)
    print *, " This ends the Ping-Pong measure."
  end subroutine Report_Performances

! private 

  function Mean(x) result (x_bar)
    implicit none
    integer, intent(in) :: x(:)
    real                :: x_bar
  ! local declarations
    integer  :: i
  ! body Mean
    x_bar = 0.0
    do i = lbound(x,1), ubound(x,1)
       x_bar = x_bar + x(i) 
    end do
    x_bar = x_bar / (ubound(x,1) - lbound(x,1) + 1.0)
  end function Mean

  function Linear_Regression(x_observed, y_observed) result (estimators)
    implicit none
    integer, intent(in) :: x_observed(:)
    integer, intent(in) :: y_observed(:)
    real                :: estimators(2) ! constant at x = 0 and slope
  ! local declarations
    integer :: i
    real :: x_bar
    real :: y_bar
    real :: slope
    real :: num
    real :: den
  ! body Linear_Regression
    if (size(x_observed,1) /= size(y_observed)) then
       print *, " Ping, Linear_Regression: mismatch in observation number."
       print *, " size of x = ", size(x_observed,1), &
               "; size of y = ", size(y_observed,1)
       estimators = 0.0
       return
    end if

    x_bar = Mean(x_observed)
    y_bar = Mean(y_observed)
    num   = 0.0
    den   = 0.0
    do i = lbound(x_observed,1), ubound(x_observed,1)
       num = num + (x_observed(i) - x_bar) * (y_observed(i) - y_bar)
       den = den + (x_observed(i) - x_bar) * (x_observed(i) - x_bar)
    end do
    slope = num / den

    estimators(1) = y_bar - slope * x_bar
    estimators(2) = slope
    return
  end function Linear_Regression
  
  subroutine Measure_Integers(max_items)
    use metrics
    use parameters
    use protocol
    implicit none

    integer, intent(in) :: max_items
  ! local declarations
    logical         :: altered     = .FALSE.
    integer         :: i           = -666
    integer         :: experiment  = -666 
    integer         :: packet_size = -666
    integer         :: going(max_items)
    integer         :: coming(max_items)
    type(durations) :: delays    = durations(-666, -666)
    type(durations) :: leaving   = durations(-666, -666)
    type(durations) :: returning = durations(-666, -666)

  ! body Measure_Integers
    ! impose value to data
    select case(data_set)
    case(1)
      going(:) = (/ (huge(i) - i,  i = 1, max_items) /)
    case default
      going(:) = (/ (i,  i = 1, max_items) /)
    end select
    ! avoid start-up effects
    call throw_packet(to      = Pong_Task_ID, &
                      packet  = going(1:PROLOGUE_PACKET_SIZE), &
                      instant = leaving)
    call catch_packet(from    = Pong_Task_ID, &
                      packet  = coming(1:PROLOGUE_PACKET_SIZE), &
                      instant = returning)

    do experiment = 1, nb_measures
      ! make a round trip
       coming(:)   = -666 ! clear with any meaningless value
       packet_size = measure(experiment)%items
       call throw_packet(to      = Pong_Task_ID, &
                         packet  = going(1:packet_size), &
                         instant = leaving)
       call catch_packet(from    = Pong_Task_ID, &
                         packet  = coming(1:packet_size), &
                         instant = returning)
       ! check the integrity of data transfer
       altered = .FALSE.
       do i = 1, packet_size
          altered = altered .OR. ( going(i) /= coming(i) )
       end do
       if (altered) then
          print *, " The round trip has corrupted the data array of integers."
          print *, " Abort the experiment unexpectedly."
          stop
       end if
       ! sum up delays within the measured time
       call receive_bouncing_delays(Pong_Task_ID, delays)
       delays%early = delays%early + 2 * chrono_latency ! only transmission
       delays%late  = delays%late  + 6 * chrono_latency ! with pack/unpack

       measure(experiment)%duration(INTEGER_TYPE)%late = &
              (returning%late - leaving%early - delays%late) / 2
       measure(experiment)%duration(INTEGER_TYPE)%early = &
              (returning%early - leaving%late - delays%early) / 2
    end do
  end subroutine Measure_Integers
  
  subroutine Measure_Reals(max_items)
    use metrics
    use parameters
    use protocol
    implicit none

    integer, intent(in) :: max_items
  ! local declarations
    logical         :: altered     = .FALSE.
    integer         :: i           = -666
    integer         :: experiment  = -666 
    integer         :: packet_size = -666
    real            :: going(max_items)
    real            :: coming(max_items)
    type(durations) :: delays    = durations(-666, -666)
    type(durations) :: leaving   = durations(-666, -666)
    type(durations) :: returning = durations(-666, -666)

  ! body Measure_Reals
    ! impose value to data
    select case(data_set)
    case(1)
       going(1) = huge(going(1))
       do i = 2, max_items
          going(i) = nearest(going(i-1), -1.0)
       end do
    case default
       going(:) = (/ (i,  i = 1, max_items) /)
    end select
    ! avoid start-up effects
    call throw_packet(to      = Pong_Task_ID, &
                      packet  = going(1:PROLOGUE_PACKET_SIZE), &
                      instant = leaving)
    call catch_packet(from    = Pong_Task_ID, &
                      packet  = coming(1:PROLOGUE_PACKET_SIZE), &
                      instant = returning)
    
    do experiment = 1, nb_measures
       ! make a round trip
       coming      = -666 ! clear with any meaningless value
       packet_size = measure(experiment)%items
       call throw_packet(to      = Pong_Task_ID, &
                         packet  = going(1:packet_size), &
                         instant = leaving)
       call catch_packet(from    = Pong_Task_ID, &
                         packet  = coming(1:packet_size), &
                         instant = returning)
       ! check the integrity of data transfer
       altered = .FALSE.
       do i = 1, packet_size
          altered = altered .OR. ( going(i) /= coming(i) )
       end do
       if (altered) then
          print *, " The round trip has corrupted the data array of integers."
          print *, " Abort the experiment unexpectedly."
          stop
       end if
       ! sum up delays within the measured time
       call receive_bouncing_delays(Pong_Task_ID, delays)
       delays%early = delays%early + 2 * chrono_latency ! only transmission
       delays%late  = delays%late  + 6 * chrono_latency ! with pack/unpack

       measure(experiment)%duration(REAL_TYPE)%late = &
              (returning%late - leaving%early - delays%late) / 2
       measure(experiment)%duration(REAL_TYPE)%early = &
              (returning%early - leaving%late - delays%early) / 2
    end do
  end subroutine Measure_Reals
  
  subroutine Measure_Complexes(max_items)
    use metrics
    use parameters
    use protocol
    implicit none

    integer, intent(in) :: max_items
  ! local declarations
    logical         :: altered     = .FALSE.
    integer         :: i           = -666
    integer         :: experiment  = -666 
    integer         :: packet_size = -666
    complex         :: going(max_items)
    complex         :: coming(max_items)
    type(durations) :: delays    = durations(-666, -666)
    type(durations) :: leaving   = durations(-666, -666)
    type(durations) :: returning = durations(-666, -666)

  ! body Measure_Complexes
    ! impose value to data
    select case(data_set)
    case(1)
       going(1) = huge(1.0)
       do i = 2, max_items
          going(i) = nearest(real(going(i-1)), -1.0)
       end do
    case default
       going(:) = (/ (i,  i = 1, max_items) /)
    end select
    ! avoid start-up effects
    call throw_packet(to      = Pong_Task_ID, &
                      packet  = going(1:PROLOGUE_PACKET_SIZE), &
                      instant = leaving)
    call catch_packet(from    = Pong_Task_ID, &
                      packet  = coming(1:PROLOGUE_PACKET_SIZE), &
                      instant = returning)
    
    do experiment = 1, nb_measures
      ! make a round trip
       coming      = -666 ! clear with any meaningless value
       packet_size = measure(experiment)%items
       call throw_packet(to      = Pong_Task_ID, &
                         packet  = going(1:packet_size), &
                         instant = leaving)
       call catch_packet(from    = Pong_Task_ID, &
                         packet  = coming(1:packet_size), &
                         instant = returning)
       ! check the integrity of data transfer
       altered = .FALSE.
       do i = 1, packet_size
          altered = altered .OR. ( going(i) /= coming(i) )
       end do
       if (altered) then
          print *, " The round trip has corrupted the data array of complexes."
          print *, " Abort the experiment unexpectedly."
          stop
       end if
       ! sum up delays within the measured time
       call receive_bouncing_delays(Pong_Task_ID, delays)
       delays%early = delays%early + 2 * chrono_latency ! only transmission
       delays%late  = delays%late  + 6 * chrono_latency ! with pack/unpack

       measure(experiment)%duration(COMPLEX_TYPE)%late = &
              (returning%late - leaving%early - delays%late) / 2
       measure(experiment)%duration(COMPLEX_TYPE)%early = &
              (returning%early - leaving%late - delays%early) / 2
    end do
  end subroutine Measure_Complexes
  
  subroutine Measure_Logicals(max_items)
    use metrics
    use parameters
    use protocol
    implicit none

    integer, intent(in) :: max_items
  ! local declarations
    logical         :: altered     = .FALSE.
    integer         :: i           = -666
    integer         :: experiment  = -666 
    integer         :: packet_size = -666
    logical         :: going(max_items)
    logical         :: coming(max_items)
    type(durations) :: delays    = durations(-666, -666)
    type(durations) :: leaving   = durations(-666, -666)
    type(durations) :: returning = durations(-666, -666)

  ! body Measure_Logicals
    ! impose value to data
    going = (/ (modulo(i,2) == 1, i = 1, max_items) /)
    ! avoid start-up effects
    call throw_packet(to      = Pong_Task_ID, &
                      packet  = going(1:PROLOGUE_PACKET_SIZE), &
                      instant = leaving)
    call catch_packet(from    = Pong_Task_ID, &
                      packet  = coming(1:PROLOGUE_PACKET_SIZE), &
                      instant = returning)
    
    do experiment = 1, nb_measures
      ! make a round trip
       coming      = (/ (modulo(i,2) == 0, i = 1, max_items) /)
                     ! clear with any meaningless value
       packet_size = measure(experiment)%items
       call throw_packet(to      = Pong_Task_ID, &
                         packet  = going(1:packet_size), &
                         instant = leaving)
       call catch_packet(from    = Pong_Task_ID, &
                         packet  = coming(1:packet_size), &
                         instant = returning)
       ! check the integrity of data transfer
       altered = .FALSE.
       do i = 1, packet_size
          altered = altered .OR. .NOT.( going(i) .eqv. coming(i) )
       end do
       if (altered) then
          print *, " The round trip has corrupted the data array of logicals."
          print *, " Abort the experiment unexpectedly."
          stop
       end if
       ! sum up delays within the measured time
       call receive_bouncing_delays(Pong_Task_ID, delays)
       delays%early = delays%early + 2 * chrono_latency ! only transmission
       delays%late  = delays%late  + 6 * chrono_latency ! with pack/unpack

       measure(experiment)%duration(LOGICAL_TYPE)%late = &
              (returning%late - leaving%early - delays%late) / 2
       measure(experiment)%duration(LOGICAL_TYPE)%early = &
              (returning%early - leaving%late - delays%early) / 2
    end do
  end subroutine Measure_Logicals
  
  subroutine Measure_Characters(max_items)
    use metrics
    use parameters
    use protocol
    implicit none

    integer, intent(in) :: max_items
  ! local declarations
    logical         :: altered     = .FALSE.
    integer         :: i           = -666
    integer         :: experiment  = -666 
    integer         :: packet_size = -666
    character       :: going(max_items)
    character       :: coming(max_items)
    type(durations) :: delays    = durations(-666, -666)
    type(durations) :: leaving   = durations(-666, -666)
    type(durations) :: returning = durations(-666, -666)

  ! body Measure_Characters
    ! impose value to data
    going = (/ (achar(32+modulo(i, 64)),  i = 1, max_items) /)
    ! avoid start-up effects
    call throw_packet(to      = Pong_Task_ID, &
                      packet  = going(1:PROLOGUE_PACKET_SIZE), &
                      instant = leaving)
    call catch_packet(from    = Pong_Task_ID, &
                      packet  = coming(1:PROLOGUE_PACKET_SIZE), &
                      instant = returning)
    
    do experiment = 1, nb_measures
      ! make a round trip
       coming      = '!' ! clear with any meaningless value
       packet_size = measure(experiment)%items
       call throw_packet(to      = Pong_Task_ID, &
                         packet  = going(1:packet_size), &
                         instant = leaving)
       call catch_packet(from    = Pong_Task_ID, &
                         packet  = coming(1:packet_size), &
                         instant = returning)
       ! check the integrity of data transfer
       altered = .FALSE.
       do i = 1, packet_size
          altered = altered .OR. ( going(i) /= coming(i) )
       end do
       if (altered) then
          print *, " The round trip has corrupted the data array of characters."
          print *, " Abort the experiment unexpectedly."
          stop
       end if
       ! sum up delays within the measured time
       call receive_bouncing_delays(Pong_Task_ID, delays)
       delays%early = delays%early + 2 * chrono_latency ! only transmission
       delays%late  = delays%late  + 6 * chrono_latency ! with pack/unpack

       measure(experiment)%duration(CHARACTER_TYPE)%late = &
              (returning%late - leaving%early - delays%late) / 2
       measure(experiment)%duration(CHARACTER_TYPE)%early = &
              (returning%early - leaving%late - delays%early) / 2
    end do
  end subroutine Measure_Characters
end program Ping
